#!/usr/bin/env python
# encoding: utf-8
#
# ** header v3.0
# This file is a part of the CaosDB Project.
#
# Copyright (C) 2018 Research Group Biomedical Physics,
# Max-Planck-Institute for Dynamics and Self-Organization Göttingen
# Copyright (C) 2020 Henrik tom Wörden
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# ** end header
#
""" Crawls a file structure and inserts Records into CaosDB based on what is
found.

CaosDB can automatically be filled with Records based on some file structure.
The Crawler will iterate over the files and test for each file whether a CFood
exists that matches the file path. If one does, it is instanciated to treat the
match. This occurs in basically three steps:
1. create a list of identifiables, i.e. unique representation of CaosDB Records
(such as an experiment belonging to a project and a date/time)
2. the identifiables are either found in CaosDB or they are created.
3. the identifiables are update based on the date in the file structure
"""


import logging
import os
import traceback
from datetime import datetime

import caosdb as db
from caosdb.exceptions import TransactionError

from .cache import Cache
from .cfood import RowCFood
from .guard import RETRIEVE
from .guard import global_guard as guard

logger = logging.getLogger(__name__)


def separated(text):
    return "-"*60 + "\n" + text


class UnknownCache(object):
    def __init__(self, interactive=False, load=False):

        if interactive and "y" == input(
                "\nDo you want to load filenames that previously were not "
                "matched by any CFood?\nIn that case, they will not show up "
                "again. (y)"):
            load = True

        if load and os.path.exists("known_cache.db"):
            with open("known_cache.db") as fi:
                self.filenames = [el.strip("\n") for el in fi.readlines()]
        else:
            self.filenames = []

    def save(self):
        with open("known_cache.db", "w") as fi:
            for name in self.filenames:
                fi.write(name + "\n")

    def add(self, el):
        self.filenames.append(el)


class Crawler(object):
    def __init__(self, cfood_types, use_cache=False,
                 abort_on_exception=True, interactive=True, hideKnown=False):
        """
        Parameters
        ----------
        cfood_types : list of CFood classes
               The Crawler will use those CFoods when crawling.
        use_cache : bool, optional
                    Whether to use caching (not re-inserting probably existing
                    objects into CaosDB), defaults to False.
        abort_on_exception : if true, exceptions are raise.
                    Otherwise the crawler continues if an exception occurs.
        interactive : boolean, optional
                      If true, questions will be posed during execution of the
                      crawl function.

        """

        self.cfood_types = cfood_types
        self.interactive = interactive
        self.report = db.Container()
        self.use_cache = use_cache
        self.hideKnown = hideKnown
        self.abort_on_exception = abort_on_exception

        if self.use_cache:
            self.cache = Cache()

    def iteritems(self):
        """ generates items to be crawled with an index"""
        yield 0, None

    def collect_cfoods(self):
        """
        This is the first phase of the crawl. It collects all cfoods that shall
        be processed. The second phase is iterating over cfoods and updating
        CaosDB. This separate first step is necessary in order to allow a
        single cfood being influenced by multiple crawled items. E.g. the
        FileCrawler can have a single cfood treat multiple files.

        This is a very basic implementation and this function should be
        overwritten by subclasses.

        The basic structure of this function should be, that what ever is
        being processed is iterated and each cfood is checked whether the
        item 'matches'. If it does, a cfood is instantiated passing the item
        as an argument.
        The match can depend on the cfoods already being created, i.e. a file
        migth no longer match because it is already treaded by an earlier
        cfood.

        should return cfoods, tbs and errors_occured.
        # TODO do this via logging?
        tbs text returned from traceback
        errors_occured True if at least one error occured
        """
        cfoods = []
        tbs = []
        errors_occured = False
        matches = {idx: [] for idx, _ in self.iteritems()}

        logger.info(separated("Matching files against CFoods"))

        for Cfood in self.cfood_types:
            logger.debug("Matching against {}...".format(Cfood.__name__))

            for idx, item in self.iteritems():
                if Cfood.match_item(item):
                    try:
                        cfoods.append(Cfood(item))
                        matches[idx].append(Cfood.__name__)
                        logger.debug("{} matched\n{}.".format(
                                Cfood.__name__,
                                item))
                    except Exception as e:
                        traceback.print_exc()
                        print(e)

                        if self.abort_on_exception:
                            raise e
                        errors_occured = True
                        tbs.append(e)

        logger.info(separated("CFoods are collecting information..."))

        for cfood in cfoods:
            cfood.collect_information()

        logger.info(separated("Trying to attach further items to created CFoods"))

        for cfood in cfoods:
            logger.debug("Matching against {}...".format(Cfood.__name__))

            for idx, item in self.iteritems():
                if cfood.looking_for(item):
                    logger.debug("{} matched\n{}.".format(
                            cfood.__class__.__name__,
                            item))
                    cfood.attach(item)
                    matches[idx].append(Cfood.__name__)

        self.check_matches(matches)

        return cfoods, tbs, errors_occured

    def check_matches(self, matches):
        # possibly load previously encountered "Missing matches" and
        # "Multiple matches"
        ucache = UnknownCache(interactive=self.interactive, load=self.hideKnown)

        for idx, item in self.iteritems():
            if len(matches[idx]) == 0:
                msg = ("ATTENTION: No matching cfood!\n"
                       "Tried to match {}\n".format(item))

                if item in ucache.filenames:
                    logger.debug(msg)
                else:
                    logger.warning(msg)
                ucache.add(item)

            if len(matches[idx]) > 1:
                msg = ("Attention: More than one matching cfood!\n"
                       + "Tried to match {}\n".format(item)
                       + "\tRecordTypes:\t" + ", ".join(
                            matches[idx])+"\n")

                if item in ucache.filenames:
                    logger.debug(msg)
                else:
                    logger.warning(msg)
                ucache.add(item)

        # Save the encountered prblem matches
        ucache.save()

    def cached_find_identifiables(self, identifiables):
        if self.use_cache:
            hashes = self.cache.update_ids_from_cache(identifiables)

        self.find_or_insert_identifiables(identifiables)

        if self.use_cache:
            self.cache.insert_list(hashes, identifiables)

    def crawl(self, security_level=RETRIEVE):
        guard.set_level(level=security_level)

        cfoods, tbs, errors_occured = self.collect_cfoods()

        if self.interactive and "y" != input("Do you want to continue? (y)"):
            return

        logger.info(separated("Creating and updating Identifiables"))

        for cfood in cfoods:
            try:
                cfood.create_identifiables()

                self.cached_find_identifiables(cfood.identifiables)

                cfood.update_identifiables()
                cfood.push_identifiables_to_CaosDB()
            except Exception as e:
                traceback.print_exc()
                print(e)

                if self.abort_on_exception:
                    raise e
                errors_occured = True
                tbs.append(e)

        if errors_occured:
            logger.warning("Crawler terminated with failures!")
            logger.warning(tbs)
        else:
            logger.info("Crawler terminated successfully!")

    # TODO remove static?
    @staticmethod
    def find_or_insert_identifiables(identifiables):
        """ Sets the ids of identifiables (that do not have already an id from the
        cache) based on searching CaosDB and retrieves those entities.
        The remaining entities (those which can not be retrieved) have no
        correspondence in CaosDB and are thus inserted.
        """
        # looking for matching entities in CaosDB when there is no valid id
        # i.e. there was none set from a cache

        for ent in identifiables:
            if ent.id is None or ent.id < 0:
                logger.debug("Looking for: {}".format(ent))
                existing = Crawler.find_existing(ent)

                if existing is not None:
                    ent.id = existing.id
            else:
                logger.debug("Id is known of: {}".format(ent))

        # insert missing, i.e. those which are not valid
        missing_identifiables = db.Container()
        missing_identifiables.extend([ent for ent in identifiables
                                      if ent.id is None or ent.id < 0])
        # TODO the following should not be necessary. Fix it

        for ent in missing_identifiables:
            ent.id = None

        if len(missing_identifiables) > 0:
            logger.info("Going to insert the following entities:")

            for ent in missing_identifiables:
                logger.info(ent)

        if len(missing_identifiables) == 0:
            logger.debug("No new entities to be inserted.")
        else:
            guard.safe_insert(missing_identifiables)

        logger.debug("Retrieving entities from CaosDB...")
        identifiables.retrieve(unique=True, raise_exception_on_error=False)

    @staticmethod
    def find_existing(entity):
        """searches for an entity that matches the identifiable in CaosDB

        Characteristics of the identifiable like, properties, name or id are
        used for the match.
        """

        if entity.name is None:
            # TODO multiple parents are ignored! Sufficient?
            query_string = "FIND Record " + entity.get_parents()[0].name
            query_string += " WITH "

            for p in entity.get_properties():
                query_string += ("'" + p.name + "'='" + str(get_value(p))
                                     + "' AND ")
            # remove the last AND
            query_string = query_string[:-4]
        else:
            query_string = "FIND '{}'".format(entity.name)

        logger.debug(query_string)
        q = db.Query(query_string)
        # the identifiable should identify an object uniquely. Thus the query
        # is using the unique keyword
        try:
            r = q.execute(unique=True)
        except TransactionError:
            r = None

        # if r is not None:
        #     print("Found Entity with id:", r.id)
        # else:
        #     print("Did not find an existing entity.")

        return r


class FileCrawler(Crawler):
    def __init__(self, files, **kwargs):
        """
        Parameters
        ----------
        files : files to be crawled

        """
        super().__init__(**kwargs)
        self.files = files

    def iteritems(self):
        for idx, p in enumerate(sorted([f.path for f in self.files])):
            yield idx, p

    @staticmethod
    def query_files(path):
        query_str = "FIND FILE WHICH IS STORED AT " + (path if path.endswith("/") else path + "/") + "**"
        logger.info("FILES QUERY: " + query_str)
        files = db.execute_query(query_str)
        logger.info("{} FILES TO BE PROCESSED.".format(len(files)))

        return files


class TableCrawler(Crawler):

    def __init__(self, table, unique_cols, recordtype, **kwargs):
        """
        Parameters
        ----------
        table : pandas DataFrame
        unique_cols : the columns that provide the properties for the
                      identifiable
        recordtype : Record Type of the Records to be created
        """
        self.table = table

        # TODO I do not like this yet, but I do not see a better way so far.
        class ThisRowCF(RowCFood):
            def __init__(self, item):
                super().__init__(item, unique_cols, recordtype)

        super().__init__(cfood_types=[ThisRowCF], **kwargs)

    def iteritems(self):
        for idx, row in self.table.iterrows():
            yield idx, row


def get_value(prop):
    """ Returns the value of a Property

    Parameters
    ----------
    prop : The property of which the value shall be returned.

    Returns
    -------
    out : The value of the property; if the value is an entity, its ID.

    """

    if isinstance(prop.value, db.Entity):
        return prop.value.id
    elif isinstance(prop.value, datetime):
        return prop.value.isoformat()
    else:
        return prop.value
